! ********************************************************************
! * GOP - Graphic Optimization Program for EHMO ICONC                *
! *                                                                  *
! * Copyright (C) 1997 by Gion Calzaferri                            *
! *                                                                  *
! * This program is free software; you can redistribute it and/or    *
! * modify it under the terms of the GNU General Public License as   *
! * published by the Free Software Foundation; Version 2 of the      *
! * License.                                                         *
! *                                                                  *
! * This program is distributed in the hope that it will be useful,  *
! * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
! * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU *
! * General Public License for more details.                         *
! *                                                                  *
! * You should have received a copy of the GNU General Public        *
! * License along with this program; if not, write to the Free       *
! * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   *
! * MA 02111-1307 USA                                                *
! ********************************************************************


!------------------------------------------------------------------------------
! PROJECT : GOP                                            by Roman Calzaferri
!------------------------------------------------------------------------------
! FILENAME      : ENERGY_FUNCTION_S.f90
! AUTHOR        : Roman Calzaferri
! DATE          : 22. August 1997
! LAST MODIFIED : 15. September 1997
! 
! VERSION S - for use with the simplex minimization routine by D. E. Shaw /
!                                                                    A. Miller
!------------------------------------------------------------------------------
! CONTENTS : 'Class' ENERGY_FUNCTION. Because fortran 90 does not really 
!            support classes and the interface for the subroutine 
!            'ENERGY_FUNCTION_CALL' is given by the minimization routine who 
!            calls it, only one instance of this 'class' is possible.
!
!            'Class' PARAM which is needed to know which properties of 
!            the .gen file correspond to the function parameters
!------------------------------------------------------------------------------
! SPECIALS:  Any one making a call to ENERGY_FUNCTION_INIT or 
!            ENERGY_FUNCTION_SPECIAL_INIT should not forget to call 
!            ENERGY_FUNCTION_CLEANUP at the end of calculation to ensure that
!            all memory is deallocated!
!------------------------------------------------------------------------------
! USED BY  : SEARCH_EMIN, file SEARCH_EMIN_S.f90
!------------------------------------------------------------------------------
! USES     : GENFILE_TYPE, file GENFILE_TYPE.f90
!            ENERTOT,      file ENERTOT.f90
!            INPUTC_ICONC, file INPUTC_ICONC.f90
!------------------------------------------------------------------------------


MODULE ENERGY_FUNCTION
   USE GENFILE_TYPE
   USE ENERTOT
   USE INPUTC_ICONC
   IMPLICIT NONE


   ! ----------------------------------- 
   TYPE PARAM
      ! used to store which vector should be updated how
      PRIVATE
         INTEGER :: WHICH_VECTOR, WHICH_PROPERTY
   END TYPE PARAM

   ! ----------------------------------- 
   ! private variables
   TYPE(GENFILE), PRIVATE                           :: MY_GENFILE
   INTEGER, PRIVATE                                 :: DIM        ! dimension
                                                                  ! of problem
   TYPE(VECTOR), DIMENSION(:), ALLOCATABLE, PRIVATE :: VECTORS
   TYPE(PARAM), DIMENSION(:), ALLOCATABLE, PRIVATE  :: MY_PARAMS
   ! ----------------------------------- 
   INTEGER                               ENERGY_FUNCTION_EVALUATIONCOUNT
                                                      ! counts how many times 
                                                      ! the function is called


   ! ----------------------------------- 
   CONTAINS

      ! ----------------------------------- 
      SUBROUTINE ENERGY_FUNCTION_INIT(GF)
         ! Initialization with a GENFILE.
         ! Parameter are taken from variations
         !------------------------------
         ! REQUIRES : GF%INITIALIZED = .TRUE.
         !------------------------------
         TYPE(GENFILE), INTENT(IN)                  :: GF
         TYPE(VARIATION), DIMENSION(:), ALLOCATABLE :: VARIATIONS
         INTEGER I

         ENERGY_FUNCTION_EVALUATIONCOUNT = 0
         MY_GENFILE = GF

         DIM = GET_VARIATION_COUNT(MY_GENFILE)

         ALLOCATE(VECTORS(GET_VECTOR_COUNT(MY_GENFILE)))
         VECTORS = GET_VECTORS(MY_GENFILE)

         ALLOCATE(VARIATIONS(DIM))
         VARIATIONS = GET_VARIATIONS(MY_GENFILE)

         ALLOCATE(MY_PARAMS(DIM))

         DO I = 1, DIM
            MY_PARAMS(I)%WHICH_VECTOR = GET_VARY_VECTOR(VARIATIONS(I))   
            MY_PARAMS(I)%WHICH_PROPERTY = GET_TYPE_OF_VAR(VARIATIONS(I))   
         END DO

         ! Minimize the calculations iconc has to make
         CALL DELETE_VARIATIONS(MY_GENFILE)
         CALL MINIMIZE_OUTPUT(MY_GENFILE)

         DEALLOCATE(VARIATIONS)
      END SUBROUTINE ENERGY_FUNCTION_INIT
      ! ----------------------------------- 

      SUBROUTINE ENERGY_FUNCTION_SPECIAL_INIT(GF)
         ! Initialization with a GENFILE.
         ! Parameter are specified by user
         !------------------------------
         ! REQUIRES : GF%INITIALIZED = .TRUE.
         !------------------------------
         TYPE(GENFILE), INTENT(IN)    :: GF
         INTEGER                         I, MAXNUM
         CHARACTER                       ANSWER

         ENERGY_FUNCTION_EVALUATIONCOUNT = 0
         MY_GENFILE = GF

         ALLOCATE(VECTORS(GET_VECTOR_COUNT(MY_GENFILE)))
         VECTORS = GET_VECTORS(MY_GENFILE)

         ! maximum number of parameters
         MAXNUM = 3*GET_VECTOR_COUNT(MY_GENFILE)

         DO 
            DO
               WRITE(*,'(A, I3, A)',ADVANCE='NO')                       &
                  ' How many parameter do you want to optimize (max. ', &
                  MAXNUM, ' )?'
               READ(*,*) DIM
               IF (DIM<=MAXNUM) EXIT
            END DO
            ALLOCATE(MY_PARAMS(DIM))

            WRITE(*,'(//,A,/)') &
               ' The specified .gen file contains the following vectors:'
            CALL PRINT_VECTOR_INFO(VECTORS)

            WRITE(*,'(/, 7(A,/))')                                          &
               ' Parameters for variation; Please choose:',               &
               '   - vector number',                                      &
               '   - Var. of 1=distance, 2=bond angle, 3=dihedral angle', &
               '  ',                                                      &
               ' Param.  vector   Var.of',                                &
               ' number  number    [1-3]',                                &
               ' -----------------------' 

            DO I=1,DIM
               WRITE(*,'(I3,A)', ADVANCE='NO') I,'     :'
               READ(*,*) MY_PARAMS(I)%WHICH_VECTOR, MY_PARAMS(I)%WHICH_PROPERTY
            END DO

            WRITE(*,*) 'Are all inputs correct(y/n)?'
            READ*, ANSWER

            IF (ANSWER/= 'n' .AND. ANSWER/='N') THEN
               EXIT
            ELSE 
               DEALLOCATE(MY_PARAMS)  !let's have a new try...
            END IF
         END DO

         ! Minimize the calculations iconc has to make
         CALL DELETE_VARIATIONS(MY_GENFILE)
         CALL MINIMIZE_OUTPUT(MY_GENFILE)

      END SUBROUTINE ENERGY_FUNCTION_SPECIAL_INIT
      ! ----------------------------------- 

      SUBROUTINE ENERGY_FUNCTION_CALL(ARGS, FUNCTION_VALUE)
         ! this is needed by the simplex routine. It calculates the energy value
         ! for a given set of parameters, using ICONC
         !------------------------------
         ! REQUIRES : INIT or SPECIAL_INIT has been called beforehand
         !------------------------------
         DOUBLE PRECISION, DIMENSION(DIM), INTENT(IN) :: ARGS
         DOUBLE PRECISION, INTENT(OUT)                :: FUNCTION_VALUE
         INTEGER                                         I, ERR

         DO I=1,DIM
            !change the vectors so they fit the arguments 'ARGS'
            SELECT CASE (MY_PARAMS(I)%WHICH_PROPERTY)
               CASE(1)
                  CALL CHANGE_DISTANCE       &
                         (VECTORS(MY_PARAMS(I)%WHICH_VECTOR), ARGS(I))
               CASE(2)
                  CALL CHANGE_BOND_ANGLE     &
                         (VECTORS(MY_PARAMS(I)%WHICH_VECTOR), ARGS(I))
               CASE(3)
                  CALL CHANGE_DIHEDRAL_ANGLE &
                         (VECTORS(MY_PARAMS(I)%WHICH_VECTOR), ARGS(I))
            END SELECT
         END DO

         !update genfile with new vectors
         CALL CHANGE_VECTORS(MY_GENFILE, VECTORS)
         CALL WRITE_TEMPFILE(MY_GENFILE)

         ! now make the calculations
         ERR = INPUTC_ICONC_EXECUTE(GET_TEMPFILE_NAME(MY_GENFILE))
         IF (ERR /= 0) THEN
            ! if this point is reached, data is inconsistent. 
            ! control cannot be returned to the main program, so we have to stop
            STOP 'Program execution stopped.'
         END IF

         !get the function value from enertot
         CALL ENERTOT_OPEN
         FUNCTION_VALUE = ENERTOT_E_VAL()
         CALL ENERTOT_CLOSE

         ENERGY_FUNCTION_EVALUATIONCOUNT = &
                               ENERGY_FUNCTION_EVALUATIONCOUNT + 1
      END SUBROUTINE ENERGY_FUNCTION_CALL
      !----------------------------------

      FUNCTION ENERGY_FUNCTION_DIMENSION
         ! returns the dimension of the problem, that is the number of parameters
         !------------------------------
         ! REQUIRES : INIT or SPECIAL_INIT has been called beforehand
         !------------------------------
         INTEGER ENERGY_FUNCTION_DIMENSION

         ENERGY_FUNCTION_DIMENSION = DIM
      END FUNCTION ENERGY_FUNCTION_DIMENSION
      !----------------------------------

      FUNCTION ENERGY_FUNCTION_CURRENT_ARGS
         ! Returns the arguments used in the last evaluation
         !------------------------------
         ! REQUIRES : INIT or SPECIAL_INIT has been called beforehand
         !------------------------------
         DOUBLE PRECISION, DIMENSION(DIM) :: ENERGY_FUNCTION_CURRENT_ARGS
         INTEGER I
         
         DO I=1,DIM
            SELECT CASE (MY_PARAMS(I)%WHICH_PROPERTY)
               CASE (1)
                  ENERGY_FUNCTION_CURRENT_ARGS(I) = &
                  GET_DISTANCE(VECTORS(MY_PARAMS(I)%WHICH_VECTOR))
               CASE (2)
                  ENERGY_FUNCTION_CURRENT_ARGS(I) = &
                  GET_BOND_ANGLE(VECTORS(MY_PARAMS(I)%WHICH_VECTOR))
               CASE (3)
                  ENERGY_FUNCTION_CURRENT_ARGS(I) = &
                  GET_DIHEDRAL_ANGLE(VECTORS(MY_PARAMS(I)%WHICH_VECTOR))
            END SELECT
         END DO
      END FUNCTION ENERGY_FUNCTION_CURRENT_ARGS
      !----------------------------------

      FUNCTION ENERGY_FUNCTION_GET_PARAMS
         !------------------------------
         ! REQUIRES : INIT or SPECIAL_INIT has been called beforehand
         !------------------------------
         TYPE(PARAM), DIMENSION(DIM) :: ENERGY_FUNCTION_GET_PARAMS
         
         ENERGY_FUNCTION_GET_PARAMS = MY_PARAMS
      END FUNCTION ENERGY_FUNCTION_GET_PARAMS   
      !----------------------------------

      SUBROUTINE ENERGY_FUNCTION_CLEANUP
         !------------------------------
         ! REQUIRES : INIT or SPECIAL_INIT has been called beforehand
         !------------------------------
         DEALLOCATE(MY_PARAMS, VECTORS)
      END SUBROUTINE ENERGY_FUNCTION_CLEANUP
      !----------------------------------
      !----------------------------------

      !routines to access params
      !----------------------------------
      FUNCTION GET_WHICH_VECTOR(P)
         TYPE(PARAM), INTENT(IN) :: P
         INTEGER GET_WHICH_VECTOR

         GET_WHICH_VECTOR = P%WHICH_VECTOR
      END FUNCTION GET_WHICH_VECTOR

      !----------------------------------
      FUNCTION GET_WHICH_PROPERTY(P)
         TYPE(PARAM), INTENT(IN) :: P
         INTEGER GET_WHICH_PROPERTY

         GET_WHICH_PROPERTY = P%WHICH_PROPERTY
      END FUNCTION GET_WHICH_PROPERTY
      !----------------------------------
      !----------------------------------

END MODULE ENERGY_FUNCTION
! -----------------------------------
! end of file
! -----------------------------------
