!------------------------------------------------------------------------------
! ********************************************************************
! * GOP - Graphic Optimization Program for EHMO ICONC                *
! *                                                                  *
! * Copyright (C) 1997 by Gion Calzaferri                            *
! *                                                                  *
! * This program is free software; you can redistribute it and/or    *
! * modify it under the terms of the GNU General Public License as   *
! * published by the Free Software Foundation; Version 2 of the      *
! * License.                                                         *
! *                                                                  *
! * This program is distributed in the hope that it will be useful,  *
! * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
! * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU *
! * General Public License for more details.                         *
! *                                                                  *
! * You should have received a copy of the GNU General Public        *
! * License along with this program; if not, write to the Free       *
! * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   *
! * MA 02111-1307 USA                                                *
! ********************************************************************


!------------------------------------------------------------------------------
! PROJECT : GOP                                            by Roman Calzaferri
!------------------------------------------------------------------------------
! FILENAME      : SEARCH_EMIN_S.f90
! AUTHOR        : Roman Calzaferri
! DATE          : 2. September 1997
! LAST MODIFIED : 15. September 1997
!
! VERSION S - for use with the simplex minimization routine by D. E. Shaw /
!                                                                    A. Miller
!------------------------------------------------------------------------------
! CONTENTS : Module SEARCH_EMIN which does coordinate the search for a minimum
!            on an energy hypersurface. Lets the user specify options and calls
!            the minimization routine. 
!            There is only one routine that can be called from outside,
!            'SEARCH_EMIN_START'. It does the whole set up and calculation.
!
!            Settings for the simplex algorithm the user can't change:
!               SIMPLEX_EXPAND = 1.E-4 ; see description for parameter 'SIMP'
!                                        in siplex routine
!               NLOOP = 2*NUMBER_OF_PARAMETERS ; the stopping rule is applied
!                                        every NLOOP evaluations
!               QUADRATIC_SURFACE = 1  ; means a quadratic surface fit is al-
!                                        ways made (parameter IQUAD in simplex
!                                        routine)
!------------------------------------------------------------------------------
! USED BY  : PROGRAM GOP,     file GOOP_MAIN.f90
!------------------------------------------------------------------------------
! USES     : SIMPLEX,         file SIMPLEX_NEW.f90
!            GENFILE_TYPE,    file GENFILE_TYPE.f90
!            ENERGY_FUNCTION, file ENERGY_FUNCTION_S.f90
!------------------------------------------------------------------------------


MODULE SEARCH_EMIN
   USE SIMPLEX
   USE GENFILE_TYPE
   USE ENERGY_FUNCTION
   IMPLICIT NONE

   !------------------------------------
   ! logical unit number for screen output
   ! this is only needed for routines DISPLAY_CURRENT_SETTINGS and
   ! PRINT_RESULTS. In this way, duplicate code can be eliminated
   INTEGER, PARAMETER :: SCREEN_OUT = 6

   !------------------------------------
   CONTAINS

     !------------------------------------
     SUBROUTINE SEARCH_EMIN_START(GF)
         !------------------------------
         ! REQUIRES : GF%INITIALIZED = .TRUE.
         !------------------------------
         TYPE(GENFILE), INTENT(IN)               :: GF
         TYPE(PARAM), DIMENSION(:), ALLOCATABLE  :: EF_PARAMS
                                            ! parameters from ENERGY_FUNCTION
         CHARACTER(LEN=20)                       :: OUTPUTFILE = 'E_MIN.OUT'
         INTEGER CHOICE, I
         DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: START_PARAMETERS
                     ! these are needed to store where the function started

            
         !Variables needed for simplex routine
         DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: &
                                                 PARAMETERS, STEP_SIZE, VAR
         DOUBLE PRECISION :: STOPPING_CRITERION, SIMPLEX_EXPAND, &
                                                 ENERGY_MINIMUM
         INTEGER :: NUMBER_OF_PARAMETERS, MAX_OF_EVALUATIONS, IPRINT, NLOOP, &
                                                 QUADRATIC_SURFACE, FAULT_FLAG



         WRITE(*, '(//, 9(A,/))') &
              '     ******************************************', &
              '     *   simplex search for energy minimum    *', &
              '     *                                        *', &
              '     *  1  use variations in .gen file to     *', &
              '     *       choose the parameters            *', &
              '     *  2  choose parameters individually     *', &
              '     *  0  exit                               *', &
              '     *                                        *', &
              '     ******************************************'

         WRITE(*, '(//, A)', ADVANCE='NO') ' option #: '
         READ*, CHOICE

         SELECT CASE (CHOICE) 
            CASE(1) 
               CALL ENERGY_FUNCTION_INIT(GF)
            CASE(2)
               CALL ENERGY_FUNCTION_SPECIAL_INIT(GF)
            CASE DEFAULT
               RETURN
         END SELECT

         ! set options for simplex routine
         CALL SET_SIMPLEX_OPTIONS

         !now the simplex algorithm is started
         CALL minim(PARAMETERS, STEP_SIZE, NUMBER_OF_PARAMETERS,  &
                    ENERGY_MINIMUM, MAX_OF_EVALUATIONS, IPRINT,   &
                    STOPPING_CRITERION, NLOOP, QUADRATIC_SURFACE, &
                    SIMPLEX_EXPAND, VAR, ENERGY_FUNCTION_CALL, FAULT_FLAG)

                                                        
         CALL PRINT_RESULTS(SCREEN_OUT)
         CALL WRITE_OUTPUTFILE

         !clean up
         DEALLOCATE(PARAMETERS, VAR, STEP_SIZE, EF_PARAMS, START_PARAMETERS)
         CALL ENERGY_FUNCTION_CLEANUP

         !------------------------------------
         ! internal procedures of search_emin_start
         CONTAINS
            SUBROUTINE SET_SIMPLEX_OPTIONS
               INTEGER D, CHOICE

               NUMBER_OF_PARAMETERS = ENERGY_FUNCTION_DIMENSION()
               D = NUMBER_OF_PARAMETERS
               ALLOCATE(PARAMETERS(D), VAR(D), STEP_SIZE(D), EF_PARAMS(D), &
                        START_PARAMETERS(D))
               EF_PARAMS = ENERGY_FUNCTION_GET_PARAMS()


               WRITE(*, '(//,7(A,/))') &
                    '     ******************************************', &
                    '     *   settings for simplex algorithm       *', &
                    '     *                                        *', &
                    '     *  1  use default settings               *', &
                    '     *  2  choose individual settings         *', &
                    '     *                                        *', &
                    '     ******************************************'
               WRITE(*, '(//, A)', ADVANCE='NO') ' option #: '
               READ*, CHOICE

               SELECT CASE(CHOICE)
                  CASE DEFAULT
                     CALL DEFAULT_SETTINGS
                  CASE(2)
                     CALL INDIVIDUAL_SETTINGS                  
               END SELECT
                    
               PARAMETERS = START_PARAMETERS

            END SUBROUTINE SET_SIMPLEX_OPTIONS
            !------------------------------------

            SUBROUTINE DEFAULT_SETTINGS
               INTEGER I
               !set options for simplex algorithm
               START_PARAMETERS = ENERGY_FUNCTION_CURRENT_ARGS()
               DO I = 1,NUMBER_OF_PARAMETERS
                  IF (GET_WHICH_PROPERTY(EF_PARAMS(I))==1) THEN
                     STEP_SIZE(I) = 0.05
                  ELSE
                     STEP_SIZE(I) = 1.
                  END IF
               END DO
              
               MAX_OF_EVALUATIONS = 1000
               IPRINT = -1

               STOPPING_CRITERION = 1.D-4
               NLOOP = 2*NUMBER_OF_PARAMETERS

               QUADRATIC_SURFACE = 1

               SIMPLEX_EXPAND = 1.D-4
            END SUBROUTINE DEFAULT_SETTINGS
            !------------------------------------

            SUBROUTINE INDIVIDUAL_SETTINGS
               INTEGER CHOICE

               CALL DEFAULT_SETTINGS  !first set up all defaults

               DO
                  CALL DISPLAY_CURRENT_SETTINGS(SCREEN_OUT)
                  WRITE(*,'(/A)', ADVANCE='NO') &
                                          ' Press ENTER to proceed to menu'
                  READ*

                  WRITE(*, '(/,12(A,/))') &
                       '     ******************************************', &
                       '     *   change :                             *', &
                       '     *                                        *', &
                       '     *  1  starting values for parameters     *', &
                       '     *  2  initial step sizes                 *', &
                       '     *  3  stopping criterion                 *', &
                       '     *  4  max. no. of function evaluations   *', &
                       '     *  5  name of output file                *', &
                       '     *  6  progress report                    *', &
                       '     *  0  exit                               *', &
                       '     *                                        *', &
                       '     ******************************************'

                  WRITE(*, '(/, A)', ADVANCE='NO') ' option #: '
                  READ*, CHOICE

                  SELECT CASE(CHOICE)
                     CASE(1)
                        CALL CHANGE_STARTING_VALUES
                     CASE(2)
                        CALL CHANGE_STEP_SIZE
                     CASE(3)
                        WRITE(*,'(//,A)', ADVANCE='NO') &
                           ' New value for stopping criterion &
                           &(should be >= 1.e-10) : '
                        READ*, STOPPING_CRITERION
                     CASE(4)
                        WRITE(*,'(//,A)', ADVANCE='NO') &
                           ' New value for maximum number of function&
                           & evaluations allowed : '
                        READ*, MAX_OF_EVALUATIONS
                     CASE(5)
                        WRITE(*, '(//,A)', ADVANCE='NO') &
                           ' Name of output file (w. extension, up to 50 &
                           &characters : '
                        READ*, OUTPUTFILE
                     CASE(6)
                        WRITE(*, '(//,A,/,A)', ADVANCE='NO') &
                           ' After how many function evaluations would you &
                           &like a report',' (neg. number=no report; 0=after &
                           &evidence of convergence) ? '
                        READ*, IPRINT
                     CASE(0)
                        EXIT   !get out of do-loop
                  END SELECT
               END DO

            END SUBROUTINE INDIVIDUAL_SETTINGS
            !------------------------------------
                 !------------------------------------
                 !routines only used in INDIVIDUAL_SETTINGS (cannot be contained)

                 SUBROUTINE CHANGE_STARTING_VALUES
                    INTEGER I

                    WRITE(*, '(///,4(A,/))') &
                       ' Type in new starting values:',                   &
                       ' ',                                               &
                       ' param.  vector   [1-3]       new ',              &
                       '  no.                        value',              &
                       ' ------  ------  ------    --------'
                    
                    DO I = 1, NUMBER_OF_PARAMETERS
                       WRITE(*, '(A,3(I4,A))', ADVANCE = 'NO')            &
                          ' ', I, '    ',                                 &
                          GET_WHICH_VECTOR(EF_PARAMS(I)),   '    ',       &
                          GET_WHICH_PROPERTY(EF_PARAMS(I)), '  :  '
                       READ(*,*) START_PARAMETERS(I)
                    END DO

                 END SUBROUTINE CHANGE_STARTING_VALUES
                 !------------------------------------

                 SUBROUTINE CHANGE_STEP_SIZE
                    INTEGER I
                    WRITE(*, '(///,4(A,/), A)')                           &
                       ' Type in new step sizes:',                        &
                       ' ',                                               &
                       ' param.  vector   [1-3]    value       new step', &
                       '  no.                                    size',   &
                       ' ------  ------  ------   ----------   --------'
                    DO I = 1, NUMBER_OF_PARAMETERS
                       WRITE(*, '(A,3(I4,A),F10.5,A)', ADVANCE = 'NO')    &
                          ' ', I, '    ',                                 &
                          GET_WHICH_VECTOR(EF_PARAMS(I)),   '    ',       &
                          GET_WHICH_PROPERTY(EF_PARAMS(I)), '    ',       &
                          START_PARAMETERS(I), '  :  '
                       READ(*,*) STEP_SIZE(I)
                    END DO
                 END SUBROUTINE CHANGE_STEP_SIZE
                 !------------------------------------
                 !------------------------------------
            !------------------------------------

            SUBROUTINE DISPLAY_CURRENT_SETTINGS(UN)
                INTEGER, INTENT(IN)  :: UN       ! logical unit number
                INTEGER I

                WRITE(UN, '(////, 4(A,/), A)')                          &
                   ' Current settings:', ' ',                           &
                   '   - starting values for parameters and initial step sizes:',&
                   '     vector  [1-3]   param. value  step size',      &
                   '     ------  ------  ------------  ---------'
                DO I= 1, NUMBER_OF_PARAMETERS
                   WRITE(UN, '(A,I5,A,I5,A,F10.5,A,F10.7)')             &
                      '     ',                                          &  
                      GET_WHICH_VECTOR(EF_PARAMS(I)), '   ',            &
                      GET_WHICH_PROPERTY(EF_PARAMS(I)), '    ',         &
                      START_PARAMETERS(I), '  ', STEP_SIZE(I)
                END DO
                WRITE(UN, '(/,A, E7.2, /, A, I10)') &
                   '   - stopping criterion : ', STOPPING_CRITERION,    &
                   '   - max. number of function evaluations : ',       &
                   MAX_OF_EVALUATIONS
                IF (IPRINT < 0) THEN
                   WRITE(UN,*) '  - progress report is disabled'
                ELSE IF (IPRINT == 0) THEN
                   WRITE(UN,*) '  - progress report after initial       &
                      &evidence of convergence'
                ELSE
                   WRITE(UN,'(A, I4, A)') &
                       '   - progress report after every ',             &
                       IPRINT, ' function evaluations'
                END IF
                          
            END SUBROUTINE DISPLAY_CURRENT_SETTINGS
            !------------------------------------

            SUBROUTINE WRITE_OUTPUTFILE
               
               OPEN(UNIT=1, FILE=OUTPUTFILE, STATUS='REPLACE', ACTION='WRITE')

               WRITE(1, '(A,A,///)') 'processed file  ', GET_GENFILE_NAME(GF)

               CALL DISPLAY_CURRENT_SETTINGS(1)
               
               CALL PRINT_RESULTS(1)
               CLOSE(1)
            END SUBROUTINE WRITE_OUTPUTFILE
            !------------------------------------

            SUBROUTINE PRINT_RESULTS(U)
               INTEGER, INTENT(IN) :: U

               WRITE(U, '(//, A, I10, A)') ' A total of ', &
                  ENERGY_FUNCTION_EVALUATIONCOUNT,  &
                  '   Function evaluations were used.'
               IF (FAULT_FLAG == 0) THEN
                  WRITE(U, '(//,A, F15.9)') ' MINIMUM = ', ENERGY_MINIMUM
                  WRITE(U, '(A)') ' FOUND AT: '
                  WRITE(U, '(/, A,/, A)')                              &
                     '     vector  [1-3]   param. value',                 &
                     '     ------  ------  ------------'
                  DO I= 1, NUMBER_OF_PARAMETERS
                     WRITE(U, '(A,I5,A,I5,A,F10.5)')                      &
                           '     ',                                          &  
                        GET_WHICH_VECTOR(EF_PARAMS(I)), '   ',            &
                        GET_WHICH_PROPERTY(EF_PARAMS(I)), '    ',         &
                        PARAMETERS(I)
                  END DO
               ELSE
                  WRITE(U,*) ' NO MINIMUM FOUND!'
                  SELECT CASE(FAULT_FLAG)
                     CASE(1)
                        WRITE(U,*) ' MAXIMUM NUMBER OF FUNCTION &
                                    &EVALUATIONS EXCEEDED'
                     CASE(2)
                        WRITE(U,*) " COULDN'T EVALUATE QUADRATIC SURFACE FITTED"
                     CASE DEFAULT
                        WRITE(U,*) ' NUMBER OF PARAMETERS < 1'
                  END SELECT
               END IF 
            END SUBROUTINE PRINT_RESULTS
            !------------------------------------

         END SUBROUTINE SEARCH_EMIN_START
         !------------------------------------
         !------------------------------------


END MODULE SEARCH_EMIN
!------------------------------------
! end of file
!------------------------------------
